#include "stdafx.h"
#include "ScriptManager.h"

#include <stdio.h>

const int max_loaded_scripts = 3;

HRESULT ScriptManager::Attach(_IBasicIdeCtlPtr basicIdeCtl, const wchar_t* root, script_loaded_status_changed_callback script_loaded_status_changed)
{
	basicIdeCtl_ = basicIdeCtl;
	root_ = root;
	script_loaded_status_changed_ = script_loaded_status_changed;

    // create the language extensions
	CComObject<CExtensions>* extensions = 0;
	HRESULT hr = CComObject<CExtensions>::CreateInstance(&extensions);
	if (FAILED(hr))
		return hr;

	extensions->AddRef();
	extensions_.Attach(extensions);
	extensions_->Attach(this);

	basicNoUIObj_.Attach(this);

	// set the secret (not needed when running under a Development Certificate)
	basicNoUIObj_->put_Secret(_bstr_t("00000000-0000-0000-0000-000000000000"));
	// initialize
	basicNoUIObj_->Initialize();
	// use design mode
	basicNoUIObj_->put_DesignModeVisible(VARIANT_TRUE);

    // allow the NoUI to be synchronized
	basicNoUIObj_->put_SynchronizedEdit(VARIANT_TRUE);
	// cause the IDE to be synchronized
	basicIdeCtl->put_SynchronizedEdit(VARIANT_TRUE);

    // add type library for COM interop
	VARIANT_BOOL result = VARIANT_FALSE;
    basicNoUIObj_->AddExtension(_bstr_t("{}"), extensions_, &result);

    // add the language extensions
    basicNoUIObj_->AddExtension(_bstr_t("-"), extensions_, &result);

    // WinWrap Basic reads/writes scripts using this class
	IVirtualFileSystemPtr virtual_file_system;
	QueryInterface(IID_IVirtualFileSystem, (void**)&virtual_file_system);
    basicNoUIObj_->putref_VirtualFileSystem(virtual_file_system);
	return hr;
}

void ScriptManager::Disconnect()
{
	VARIANT_BOOL bResult = VARIANT_FALSE;
	basicNoUIObj_->Disconnect(&bResult);
}

void ScriptManager::EnterDesignMode(VARIANT_BOOL* pCancel)
{
    if (active_script_node_)
        *pCancel = VARIANT_TRUE; // executing in a script, can't unload
    else
    {
        // unload every loaded script
        for (auto script_node : loaded_script_nodes_)
        {
            script_node->Unload();
			script_loaded_status_changed_(script_node->ScriptName(), false);
        }

		// no scripts are loaded
        loaded_script_nodes_.clear();
    }
}

HRESULT ScriptManager::EvaluateScript(ScriptNode* script_node, VARIANT& result, const VARIANT* default_value)
{
    // script is active
	ScriptNode* previous_active_script_node = active_script_node_;
    active_script_node_ = script_node;
    // a run-time error will be return failed hr
    HRESULT hr = script_node->Evaluate(result, default_value);
    // script is no longer active
    active_script_node_ = previous_active_script_node;

	return hr;
}

void ScriptManager::EvaluateScripts(const std::list<std::wstring>& script_names, std::list<variant_t>& results, const VARIANT* default_value)
{
	HRESULT hr = S_OK;
	for (const auto& script_name : script_names)
	{
        // load the script, a syntax error will return a failed hr
		ScriptNode* script_node = 0;
		hr = LoadScript(script_name.c_str(), script_node);
		if (FAILED(hr))
			break;

        // evaluate the script, a run-time error will return a failed hr
		variant_t result;
		hr = EvaluateScript(script_node, result, default_value);
		if (FAILED(hr))
			break; // error

		results.push_back(result);
	}

	if (FAILED(hr))
	{
		CComPtr<IErrorInfo> pErrorInfo;
		if (SUCCEEDED(GetErrorInfo(0, &pErrorInfo)))
		{
			// should always happen
			// create ErrObject from IErrorInfo 
			CComPtr<IDispatch> Err = new CBasicErrorObj(hr, pErrorInfo);
			if (hr != 0x800a27da)
			{
				// report the syntax or run-time error to the remote
				VARIANT_BOOL bResult = VARIANT_FALSE;
				basicNoUIObj_->ReportError(Err, &bResult);
			}

			CComDispatchDriver dd = Err;
			variant_t v;
			dd.GetPropertyByName(L"Description", &v);
			v.ChangeType(VT_BSTR);
			std::wstring description = v.bstrVal;
			if (description[0] == L'(')
			{
				size_t pos = description.find(L')');
				if (pos != std::string::npos)
				{
					while (description[++pos] == L' ')
						;

					description.erase(0,pos);
				}
			}

			size_t pos = description.find(L"\r\n");
			if (pos != std::string::npos)
				description.erase(pos);

			description.insert(0,L"<exception: ");
			description += L">";
			v = description.c_str();
			results.push_back(v);
		}
	}
}

HRESULT ScriptManager::LoadScript(const wchar_t* script_name, ScriptNode*& script_node)
{
	std::wstring name = script_name;
	if (name[0] != L'\\')
	{
        // relative script path
        if (!active_script_node_)
            name = L"\\" + name;
        else
		{
			bstr_t namex(name.c_str());
			Combine(bstr_t(active_script_node_->ScriptName().c_str()), bstr_t(name.c_str()), &namex.GetBSTR());
			name = namex;
		}
	}

	script_node = 0;
	// is the script alread loaded
	auto it = std::find_if(loaded_script_nodes_.begin(), loaded_script_nodes_.end(), [&name](auto script_nodex) { return script_nodex->ScriptName() == name; });
	if (it != loaded_script_nodes_.end())
		script_node = *it;

	if (!script_node)
		// not active or loaded
		script_node = new ScriptNode(name.c_str());

    // evaluate a script node
    if (!script_node->IsLoaded())
    {
        // load the script, a syntax error will return a failed hr
        HRESULT hr = script_node->Load(basicNoUIObj_, fast_);
		if (FAILED(hr))
			return hr; // syntax error

		// add to front of the loaded list
		loaded_script_nodes_.push_front(script_node);
		// update tree view with new status
		script_loaded_status_changed_(script_node->ScriptName(), true);

		while (loaded_script_nodes_.size() > max_loaded_scripts)
		{
			// remove the back (oldest loaded script node)
			auto oldest_script_node = loaded_script_nodes_.back();
			if (oldest_script_node->IsActive())
				break;

			loaded_script_nodes_.pop_back();
			oldest_script_node->Unload();
			// update tree view with new status
			script_loaded_status_changed_(oldest_script_node->ScriptName(), false);
		}
    }
	else if (script_node != loaded_script_nodes_.front())
	{
		// move to the front of the loaded list
		loaded_script_nodes_.remove(script_node);
		loaded_script_nodes_.push_front(script_node);
	}

	return S_OK;
}

bool ScriptManager::Shutdown()
{
	short iResult = 0;
	basicNoUIObj_->Shutdown(&iResult);
	return iResult == 0;
}

void ScriptManager::SynchronizeIde(BSTR* Param, long Id)
{
	// synchronize the IDE with the NoUI 
	basicIdeCtl_->Synchronize(Param, Id);
}

void ScriptManager::SynchronizeNoUI(BSTR* Param, long Id)
{
	// synchronize the NoUI with the IDE
	basicNoUIObj_->Synchronize(Param, Id);
}

HRESULT __stdcall ScriptManager::GetCaption(BSTR Name, BSTR* pretval) 
{
	std::wstring name = Name;
	if (name != L"<none>")
	{
		size_t x = name.rfind(L'\\');
		if (x == (size_t)-1)
			x = 0;

		size_t y = name.find(L'.', x);
		if (y == (size_t)-1)
			y = name.size();

		name = name.substr(x, y-x);
	}

	*pretval = SysAllocString(name.c_str());
	return S_OK;
}

HRESULT __stdcall ScriptManager::GetTimeStamp(BSTR Name, DATE * pretval ) 
{
	return E_NOTIMPL;
}

HRESULT __stdcall ScriptManager::Combine(BSTR BaseName, BSTR Name, BSTR * pretval ) 
{
	std::wstring name = Name;
	if (name == L"<none>")
	{
		*pretval = SysAllocString(L"<none>");
		return S_OK;
	}

    // determine the root directory
	std::wstring root = root_;
	if (SysStringLen(BaseName))
	{
		root += BaseName;
		size_t pos = root.rfind(L'\\');
		if (pos != std::wstring::npos)
			root.erase(pos);
	}

    // script names that start with \ are absolute paths
	if (name[0] != L'\\')
		name.insert(0, L"\\");

	// create path
	name = root + name;
	wchar_t szPath[MAX_PATH];
	GetFullPathName(name.c_str(), MAX_PATH, szPath, 0);
	name = szPath;

	// insure that the name is still within the one of the root_ sub directories
    if (name.substr(0,root_.size()) != root_ || name[root_.size()] != L'\\')
	{
		*pretval = SysAllocString(L"<none>");
		return S_OK;
	}

    // this is the script name
	*pretval = SysAllocString(name.c_str()+root_.size());
	return S_OK;
}

HRESULT __stdcall ScriptManager::Exists(BSTR Name, VARIANT_BOOL * pretval) 
{
	std::wstring name = root_;
	name += Name;
	*pretval = PathFileExists(name.c_str()) ? VARIANT_TRUE : VARIANT_FALSE;
	return S_OK;
}

HRESULT __stdcall ScriptManager::Read(BSTR Name, BSTR * pretval) 
{
	if (!wcscmp(Name, L"<none>"))
	{
		*pretval = SysAllocString(L"'#ReadOnly");
		return S_OK;
	}

	std::wstring name = root_;
	name += Name;
    HANDLE hFile = CreateFile(name.c_str(), GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (hFile == INVALID_HANDLE_VALUE)
		return MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WIN32, GetLastError());

    DWORD dwFileSize = GetFileSize(hFile, &dwFileSize);
    BYTE* data = new BYTE[dwFileSize];
    DWORD read = -1;
	HRESULT hr = S_OK;
    if (!ReadFile(hFile, data, dwFileSize, &read, NULL))
		hr = MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WIN32, GetLastError());

	CloseHandle(hFile);
	if (FAILED(hr))
	{
		delete [] data;
		return hr;
	}

	bool unicode = false;
	if (dwFileSize >= 2)
	{
		wchar_t c;
		memcpy(&c,data,2);
		if (c == L'\xfeff')
		{
			// little-endian on little-endian machine
			// big-endian on big-endian machine
			unicode = true;
			dwFileSize -= 2;
		}
		else if (c == L'\xfffe')
		{
			// little-endian on big-endian machine
			// big-endian on little-endian machine
			unicode = true;
			dwFileSize -= 2;
			// swap bytes
			for (DWORD i = 0; i < dwFileSize; i += 2)
			{
				BYTE b = data[i];
				data[i] = data[i+1];
				data[i+1] = b;
			}
		}
	}

	if (!unicode)
	{
		UINT cp = CP_ACP;
		BYTE* p = data;
		if (dwFileSize >= 3 && p[0] == 0xef && p[1] == 0xbb && p[2] == 0xbf)
		{
			// UTF8 BOM
			cp = CP_UTF8;
			p += 3;
			dwFileSize -= 3;
		}

		int len = MultiByteToWideChar(cp, 0, (LPCCH)p, dwFileSize, 0, 0);
		BYTE* data2 = new BYTE[len*2];
		MultiByteToWideChar(cp, 0, (LPCCH)p, dwFileSize, (LPWSTR)data2, len);
		delete [] data;
		data = data2;
		dwFileSize = len*2;
	}

	*pretval = SysAllocStringByteLen((LPCSTR)data, dwFileSize);
	return S_OK;
}

HRESULT __stdcall ScriptManager::Write(BSTR Name, BSTR text) 
{
	std::wstring name = root_;
	name += Name;
    HANDLE hFile = CreateFile(name.c_str(), GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
    if (hFile == INVALID_HANDLE_VALUE)
		return MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WIN32, GetLastError());

	BOOL bUsedDefaultChar = FALSE;
	int len = 0;
	if (text)
		len = WideCharToMultiByte(CP_UTF8, 0, text, SysStringLen(text), 0, 0, NULL, &bUsedDefaultChar);

	BYTE* data = new BYTE[len+3];
	data[0] = 0xef;
	data[1] = 0xbb;
	data[2] = 0xbf;
	if (text)
		WideCharToMultiByte(CP_UTF8, 0, text, SysStringLen(text), (LPSTR)data+3, len, NULL, &bUsedDefaultChar);

	HRESULT hr = S_OK;
	DWORD dwBytesWritten = 0;
	if (!WriteFile(hFile, data, len+3, &dwBytesWritten, NULL))
		hr = MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WIN32, GetLastError());

	CloseHandle(hFile);
	return hr;
}

HRESULT __stdcall ScriptManager::Delete(BSTR Name) 
{
	return E_NOTIMPL;
}
