#pragma once

#include "BasicAtl.h"

typedef CComPtr<Handler> HandlerPtr_;
typedef CComPtr<FastEvaluate0> FastEvaluate0Ptr_;

class ScriptNode
{
	std::wstring script_name_;
	std::wstring proc_name_;
	// Module containing executable script code
	CComDispatchDriver module_;
	// Handler validates the function prototype during parsing and provides evaluation at run-time
	HandlerPtr_ handler_;
	// FastEvaluate0 provides fast evaluation at run-time
	FastEvaluate0Ptr_ fast_evaluate_;
	// active count
	int active_count_;

public:
	void Active(bool locking) { active_count_ += locking ? 1 : -1; }
	bool IsLoaded() const { return handler_ != 0; }
	bool IsActive() const { return active_count_ > 0; }
	const std::wstring& ProcName() const { return proc_name_; }
	const std::wstring& ScriptName() const { return script_name_; }

	ScriptNode(const wchar_t* script_name) :
		script_name_(script_name),
		active_count_(0)
	{
		size_t x = script_name_.rfind(L'\\')+1;
		size_t y = script_name_.find(L'.', x);
		if (y == (size_t)-1)
			y = script_name_.size();

        // proc name is the script name's final part without the extension
		proc_name_ = script_name_.substr(x, y-x);
	}

	static bool ismissing(const VARIANT* pvar)
	{
		if (pvar)
			switch (pvar->vt)
			{
			case VT_ERROR:
				return pvar->scode == DISP_E_PARAMNOTFOUND;
			case VT_BYREF|VT_ERROR:
				return *pvar->pscode == DISP_E_PARAMNOTFOUND;
			case VT_BYREF|VT_VARIANT:
				return ismissing(pvar->pvarVal);
			}

		return false;
	}

	HRESULT Evaluate(VARIANT& result, const VARIANT* default_value = 0)
	{
		HRESULT hr = S_OK;
		VARIANT_BOOL bExists = VARIANT_FALSE;
		handler_->get_Exists(&bExists);
		if (bExists)
		{
			// the handler exists if the script defines Function <ProcName>() As Variant
			if (fast_evaluate_)
				hr = fast_evaluate_->Evaluate(&result);
			else
			{
				LPSAFEARRAY psa = 0;
				hr = handler_->Evaluate(&psa, &result);
			}
		}
		else if (!ismissing(default_value))
			result = *default_value;

		return hr; // success
	}

	HRESULT Load(CBasicNoUIObj& basic, bool fast)
	{
        // create a handler for accessing the sub
		std::wstring prototype = script_name_;
		prototype += L"|Function ";
		prototype += proc_name_;
		prototype += L"() As Variant";
        basic->CreateHandler(bstr_t(prototype.c_str()), &handler_);
        // load the module -- public symbols are only accessible via module_
		basic->ModuleInstance(bstr_t(script_name_.c_str()), VARIANT_FALSE, &module_);
        if (!module_)
		{
			// syntax error
			handler_.Release();
			// create SetErrorInfo for the syntax error
			return basic.SetErrorInfo();
		}

		if (fast)
		{
			CComPtr<Fast> fastx;
			handler_->get_Fast(&fastx);
			fastx.QueryInterface(&fast_evaluate_);
		}

		return S_OK; // success

	}

	void Unload()
	{
		if (module_)
			module_.Release();

		if (handler_)
			handler_.Release();

		if (fast_evaluate_)
			fast_evaluate_.Release();
	}
};
